import json
import re


# TODO: Test this class explicitly, the implicit testing we got now
# works but it doesn't define the behavior.
class Response(object):
    """A nicely wrapped HTTP Response

    This class will be instantiated from API calls and is not intended
    to be used in any other way.

    Special handling of the responses:

    1. A response has a boolean value that corresponds with :meth:`is_ok`:

        >>> bool(Response(404, 'Not found'))
        False
        >>> bool(Response(200, 'OK'))
        True

    2. If the payload is json you can access the json object through
       the response as if it was a `dict`:
        >>> response = Response(200, '{"hello": "there"}',
                                {'Content-Type': 'application/json'})
        >>> response['hello']
        "there"

    3. If the payload is json you can look if a key is part of the
       payload as you would with a `dict`:
        >>> response = Response(200, '{"hello": "there"}',
                                {'Content-Type': 'application/json'})
        >>> 'hello' in response
        True

    Args:
      status_code (int): The HTTP status of this response
      body (str, fp): The HTTP response body/payload of this response
      headers (dict, optional): The headers supplied by this response
      ok_status (int, optional): If the `status_code` is this then this
        response is successful. Default: 200
    """
    def __init__(self, status_code, body, headers=None, ok_status=200):
        self.__body = None
        self.status_code = status_code
        self._body = body
        self._body_parsed = None
        self.content_type = headers.get('content-type', '').split(';')[0] or False
        self.headers = headers or {}
        self.ok_status = ok_status or 200

    @property
    def is_ok(self):
        """Whether this response is considered successful

        Returns
          bool: True if `status_code` is `ok_status`
        """
        return self.status_code == self.ok_status

    @property
    def is_json(self):
        """
        Returns:
          bool: True if `content_type` is `application/json`
        """
        return self.content_type.startswith('application/json') or \
            re.match(r'application/vnd.go.cd.v(\d+)\+json', self.content_type)

    def __bool__(self):
        # XXX
        # I'm not sure if this is a good idea or not,
        # but I think a response should be true/false depending on whether it was ok or not.
        # Let's try it out and see whether it's a crazy idea.
        return self.is_ok
    __nonzero__ = __bool__

    def __getitem__(self, item):
        if self.is_json:
            return self.payload[item]

        self._raise_non_json_response()

    def __contains__(self, item):
        if self.is_json:
            return item in self.payload

        self._raise_non_json_response()

    @property
    def payload(self):
        """
        Returns:
          `str` when not json.
          `dict` when json.
        """
        if self.is_json:
            if not self._body_parsed:
                if hasattr(self._body, 'decode'):
                    body = self._body.decode('utf-8')
                else:
                    body = self._body

                self._body_parsed = json.loads(body)

            return self._body_parsed
        else:
            return self._body
    #: Alias for :meth:`payload`
    body = payload

    @property
    def _body(self):
        if hasattr(self.__body, 'read'):
            self.__body = self.__body.read()

        return self.__body

    @_body.setter
    def _body(self, value):
        self.__body = value

    @property
    def fp(self):
        """Returns a file-like object if the class was instantiated with one

        Returns:
          None, file-like object: If :attribute:`_body` responds to read else None
        """
        if hasattr(self.__body, 'read'):
            return self.__body

        return None

    @property
    def etag(self):
        return self.headers["ETag"]

    @classmethod
    def _from_request(cls, response, ok_status=None):
        return Response(
            response.code,
            response,
            response.headers,
            ok_status=ok_status
        )

    @classmethod
    def _from_http_error(cls, http_error):
        return Response(
            http_error.code,
            http_error.fp.read(),
            http_error.headers,
        )

    @classmethod
    def _from_json(cls, body):
        return Response(200, json.dumps(body), {'content-type': 'application/json'})

    def _raise_non_json_response(self):
        raise AttributeError(
            "Can't lookup item in a non-JSON response.",
            content_type=self.content_type
        )



import re
import json

try:
    # python2
    from urlparse import urljoin
    from urllib2 import (
        urlopen,
        HTTPPasswordMgrWithDefaultRealm,
        HTTPBasicAuthHandler,
        HTTPHandler,
        HTTPSHandler,
        install_opener,
        build_opener,
        Request,
    )

    class CustomRequest(Request):
        """Python2's Request class has no support
        for setting the HTTP method."""

        def __init__(self, *args, **kargs):
            if "method" in kargs:
                self.__method = kargs["method"]
                del kargs["method"]
            else:
                self.__method = None
            Request.__init__(self, *args, **kargs)

        def get_method(self):
            if self.__method is not None:
                return self.__method
            else:
                return Request.get_method(self)

except ImportError:  # pragma: no cover
    # python3
    from urllib.parse import urljoin
    from urllib.request import (
        urlopen,
        HTTPPasswordMgrWithDefaultRealm,
        HTTPBasicAuthHandler,
        HTTPHandler,
        HTTPSHandler,
        install_opener,
        build_opener,
        Request,
    )

    CustomRequest = Request

from gocd.api import Pipeline, PipelineGroups, Stage

__all__ = ['Server', 'AuthenticationFailed']


class AuthenticationFailed(Exception):
    pass


class Server(object):
    """Interacting with the Go server

    If user and password is supplied the client will try to login using
    HTTP Basic Auth on each request.

    The intention is to use this class as a jumping off point to the
    nicer API wrappers in the :mod:`gocd.api` package.

    Example of intended interaction with this class::

      >>> import gocd
      >>> go_server = gocd.Server('http://localhost:8153', 'admin', 'badger')
      >>> pipeline = go_server.pipeline('up42')
      >>> response = pipeline.pause('Admin says no work for you.')
      >>> response.is_ok
      True

    Args:
      host (str): The base URL for your go server.
        Example: http://go.example.com/
      user (str): The username to login as
      password (str): The password for this user
    """
    SESSION_COOKIE_NAME = 'JSESSIONID'

    #: Sets the debug level for the urllib2 HTTP(s) handlers
    request_debug_level = 0

    _session_id = None
    _authenticity_token = None

    def __init__(self, host, user=None, password=None):
        self.host = host
        self.user = user
        self.password = password

        if self.user and self.password:
            self._add_basic_auth()

    def get(self, path):
        """Performs a HTTP GET request to the Go server

        Args:
          path (str): The full path on the Go server to request.
            This includes any query string attributes.

        Raises:
          HTTPError: when the HTTP request fails.

        Returns:
          file like object: The response from a
            :func:`urllib2.urlopen` call
        """
        return self.request(path, method='GET')

    def post(self, path, **post_args):
        """Performs a HTTP POST request to the Go server

        Args:
          path (str): The full path on the Go server to request.
            This includes any query string attributes.
          **post_args: Any POST arguments that should be sent to the server

        Raises:
          HTTPError: when the HTTP request fails.

        Returns:
          file like object: The response from a
            :func:`urllib2.urlopen` call
        """
        return self.request(path, data=post_args, method='POST')

    def request(self, path, data=None, headers=None, method=None):
        """Performs a HTTP request to the Go server

        Args:
          path (str): The full path on the Go server to request.
            This includes any query string attributes.
          data (str, dict, bool, optional): If any data is present this
            request will become a POST request.
          headers (dict, optional): Headers to set for this particular
            request

        Raises:
          HTTPError: when the HTTP request fails.

        Returns:
          file like object: The response from a
            :func:`urllib2.urlopen` call
        """
        if isinstance(data, str):
            data = data.encode('utf-8')
        response = urlopen(self._request(path, data=data, headers=headers, method=method))
        self._set_session_cookie(response)

        return response

    def add_logged_in_session(self, response=None):
        """Make the request appear to be coming from a browser

        This is to interact with older parts of Go that doesn't have a
        proper API call to be made. What will be done:

        1. If no response passed in a call to `go/api/pipelines.xml` is
           made to get a valid session
        2. `JSESSIONID` will be populated from this request
        3. A request to `go/pipelines` will be so the
           `authenticity_token` (CSRF) can be extracted. It will then
           silently be injected into `post_args` on any POST calls that
           doesn't start with `go/api` from this point.

        Args:
          response: a :class:`Response` object from a previously successful
            API call. So we won't have to query `go/api/pipelines.xml`
            unnecessarily.

        Raises:
          HTTPError: when the HTTP request fails.
          AuthenticationFailed: when failing to get the `session_id`
            or the `authenticity_token`.
        """
        if not response:
            response = self.get('go/api/pipelines.xml')

        self._set_session_cookie(response)

        if not self._session_id:
            raise AuthenticationFailed('No session id extracted from request.')

        response = self.get('go/pipelines')
        match = re.search(
            r'name="authenticity_token".+?value="([^"]+)',
            response.read().decode('utf-8')
        )
        if match:
            self._authenticity_token = match.group(1)
        else:
            raise AuthenticationFailed('Authenticity token not found on page')

    def _set_session_cookie(self, response):
        if 'set-cookie' not in response.headers:
            return

        for cookie in response.headers['set-cookie'].split(';'):
            if cookie.startswith(self.SESSION_COOKIE_NAME):
                self._session_id = cookie

    def pipeline(self, name):
        """Instantiates a :class:`Pipeline` with the given name.

        Args:
          name: The name of the pipeline you want to interact with

        Returns:
          Pipeline: an instantiated :class:`Pipeline`.
        """
        return Pipeline(self, name)

    def pipeline_groups(self):
        """Returns an instance of :class:`PipelineGroups`

        Returns:
          PipelineGroups: an instantiated :class:`PipelineGroups`.
        """
        return PipelineGroups(self)

    def stage(self, pipeline_name, stage_name, pipeline_counter=None):
        """Returns an instance of :class:`Stage`

        Args:
            pipeline_name (str): Name of the pipeline the stage belongs to
            stage_name (str): Name of the stage to act on
            pipeline_counter (int): The pipeline instance the stage is for.

        Returns:
          Stage: an instantiated :class:`Stage`.
        """
        return Stage(self, pipeline_name, stage_name, pipeline_counter=pipeline_counter)

    def _add_basic_auth(self):
        auth_handler = HTTPBasicAuthHandler(
            HTTPPasswordMgrWithDefaultRealm()
        )
        auth_handler.add_password(
            realm=None,
            uri=self.host,
            user=self.user,
            passwd=self.password,
        )
        install_opener(build_opener(
            auth_handler,
            HTTPHandler(debuglevel=self.request_debug_level),
            HTTPSHandler(debuglevel=self.request_debug_level),
        ))

    def _request(self, path, data=None, headers=None, method=None):
        default_headers = {'User-Agent': 'py-gocd'}
        if self._session_id:
            default_headers['Cookie'] = self._session_id
        default_headers.update(headers or {})

        data = self._inject_authenticity_token(data, path)
        return CustomRequest(
            self._url(path),
            data=self._encode_data(data),
            headers=default_headers,
            method=method,
        )

    def _encode_data(self, data):
        if isinstance(data, dict):
            return json.dumps(data).encode('utf-8')
        elif isinstance(data, str):
            return data.encode('utf-8')
        elif isinstance(data, bytes):
            return data
        elif data is True:
            return ''.encode('utf-8')
        else:
            return None

    def _url(self, path):
        return urljoin(self.host, path)

    def _inject_authenticity_token(self, data, path):
        if (data is None or not self._authenticity_token or path.startswith('go/api')):
            return data

        if data == '':
            data = {}

        data.update(authenticity_token=self._authenticity_token)
        return data



try:
    # python2
    from urllib2 import HTTPError
except ImportError:  # pragma: no cover
    # python3
    from urllib.request import HTTPError


from gocd.api.response import Response


class Endpoint(object):
    # The cached responses
    _base_path = None
    _id = None

    id = None  #: This points to an attribute that contains the id of this endpoint
    base_path = None  #: URI base path to search for. Any URL generated will start with this.
    server = None  #: An instance of :class:`gocd.Server`.

    def get_id(self):
        if self._id is None:
            if self.id is not None:
                self._id = getattr(self, self.id, None)

            if self._id is None:
                raise NotImplementedError(
                    'id is not set. It is needed when calling a specific '
                    'instance of an endpoint. '
                )

        return self._id

    def get_base_path(self):
        if getattr(self, 'base_path', None) is None:
            raise NotImplementedError(
                "base_path is not set. I don't know where to query on "
                "the Go server."
            )

        if not self._base_path:
            self._base_path = self.base_path.format(id=self.get_id())

        return self._base_path

    def _join_path(self, path):
        return '{0}/{1}'.format(self.get_base_path(), path).replace('//', '/')

    def _get(self, path, ok_status=None, headers=None):
        return self._request(path, ok_status=ok_status, headers=headers, method="GET")

    def _post(self, path, ok_status=None, headers=None, **post_args):
        return self._request(path, ok_status=ok_status, data=post_args or None,
                             headers=headers, method="POST")

    # TODO: Add tests for adding headers
    def _request(self, path, ok_status, data=None, headers=None, method=None):
        try:
            return Response._from_request(
                self.server.request(
                    self._join_path(path),
                    data=data,
                    headers=headers,
                    method=method,
                ),
                ok_status=ok_status
            )
        except HTTPError as exc:
            return Response._from_http_error(exc)



from .pipeline import Pipeline
from .pipeline_groups import PipelineGroups
from .pipeline_config import PipelineConfig
from .artifact import Artifact
from .stage import Stage
from .pluggable_scm import PluggableSCM
from .template_config import TemplateConfig

__all__ = ['Pipeline', 'PipelineGroups', 'Artifact', 'Stage',
           'PipelineConfig', 'PluggableSCM', 'TemplateConfig']



try:
    #python2
    from urllib import urlencode as _urlencode
except ImportError:
    #python3
    from urllib.parse import urlencode as _urlencode


def flatten(d):
    """Return a dict as a list of lists.

    >>> flatten({"a": "b"})
    [['a', 'b']]
    >>> flatten({"a": [1, 2, 3]})
    [['a', [1, 2, 3]]]
    >>> flatten({"a": {"b": "c"}})
    [['a', 'b', 'c']]
    >>> flatten({"a": {"b": {"c": "e"}}})
    [['a', 'b', 'c', 'e']]
    >>> flatten({"a": {"b": "c", "d": "e"}})
    [['a', 'b', 'c'], ['a', 'd', 'e']]
    >>> flatten({"a": {"b": "c", "d": "e"}, "b": {"c": "d"}})
    [['a', 'b', 'c'], ['a', 'd', 'e'], ['b', 'c', 'd']]

    """

    if not isinstance(d, dict):
        return [[d]]

    returned = []
    for key, value in d.items():
        # Each key, value is treated as a row.
        nested = flatten(value)
        for nest in nested:
            current_row = [key]
            current_row.extend(nest)
            returned.append(current_row)

    return returned


def parametrize(params):
    """Return list of params as params.

    >>> parametrize(['a'])
    'a'
    >>> parametrize(['a', 'b'])
    'a[b]'
    >>> parametrize(['a', 'b', 'c'])
    'a[b][c]'

    """
    returned = str(params[0])
    returned += "".join("[" + str(p) + "]" for p in params[1:])
    return returned


def urlencode(params):
    """Urlencode a multidimensional dict."""

    # Not doing duck typing here. Will make debugging easier.
    if not isinstance(params, dict):
        raise TypeError("Only dicts are supported.")

    params = flatten(params)

    url_params = {}
    for param in params:
        value = param.pop()

        name = parametrize(param)
        if isinstance(value, (list, tuple)):
            name += "[]"

        url_params[name] = value

    return _urlencode(url_params, doseq=True)



from gocd.api.endpoint import Endpoint

__all__ = ['Stage']


class Stage(Endpoint):
    base_path = 'go/api/stages/{id}'

    def __init__(self, server, pipeline_name, stage_name, pipeline_counter=None):
        """A wrapper for the `Go stage API`__

        .. __: http://api.go.cd/current/#stages

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          pipeline_name (str): The name of the pipeline we're working on
          stage_name (str): The name of the stage we're working on
        """
        self.server = server
        self.pipeline_name = pipeline_name
        self.pipeline_counter = pipeline_counter
        self.stage_name = stage_name

    def get_id(self):
        return "{pipeline}/{stage}".format(pipeline=self.pipeline_name,
                                           stage=self.stage_name)

    def cancel(self):
        """Cancels a currently running stage

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._post('/cancel', headers={"Confirm": True})

    def history(self, offset=0):
        """Lists previous instances/runs of the stage

        See the `Go stage history documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-stage-history

        Args:
          offset (int, optional): How many instances to skip for this response.

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get('/history/{offset:d}'.format(offset=offset or 0))

    def instance(self, counter=None, pipeline_counter=None):
        """Returns all the information regarding a specific stage run

        See the `Go stage instance documentation`__ for examples.

        .. __: http://api.go.cd/current/#get-stage-instance

        Args:
          counter (int): The stage instance to fetch.
            If falsey returns the latest stage instance from :meth:`history`.
          pipeline_counter (int): The pipeline instance for which to fetch
            the stage. If falsey returns the latest pipeline instance.

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        pipeline_counter = pipeline_counter or self.pipeline_counter
        pipeline_instance = None

        if not pipeline_counter:
            pipeline_instance = self.server.pipeline(self.pipeline_name).instance()
            self.pipeline_counter = int(pipeline_instance['counter'])

        if not counter:
            if pipeline_instance is None:
                pipeline_instance = (
                    self.server
                        .pipeline(self.pipeline_name)
                        .instance(pipeline_counter)
                )

            for stages in pipeline_instance['stages']:
                if stages['name'] == self.stage_name:
                    return self.instance(
                        counter=int(stages['counter']),
                        pipeline_counter=pipeline_counter
                    )

        return self._get('/instance/{pipeline_counter:d}/{counter:d}'
                         .format(pipeline_counter=pipeline_counter, counter=counter))



import json
from gocd.api.endpoint import Endpoint

__all__ = ['TemplateConfig']


class TemplateConfig(Endpoint):
    base_path = 'go/api/admin/templates'
    id = 'name'

    def __init__(self, server, name, api_version=4):
        """A wrapper for the `Go template config API`__

        .. __: https://api.go.cd/current/#template-config

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          name (str): The name of the template we're working on
        """
        self.server = server
        self.name = name
        self.api_version = api_version

    def get(self):
        """Get template config for specified template name.

        See `The template config object`__ for example responses.

        .. __: https://api.go.cd/current/#the-template-config-object

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get(self.name, headers={"Accept": self._accept_header_value})

    def edit(self, config, etag):
        """Update template config for specified template name.

        .. __: https://api.go.cd/current/#edit-template-config

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        data = self._json_encode(config)
        headers = self._default_headers()

        if etag is not None:
            headers["If-Match"] = etag

        return self._request(self.name,
                             ok_status=None,
                             data=data,
                             headers=headers,
                             method="PUT")

    def create(self, config):
        """Create template config for specified template name.

        .. __: https://api.go.cd/current/#create-template-config

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        assert config["name"] == self.name, "Given config is not for this template"

        data = self._json_encode(config)
        headers = self._default_headers()

        return self._request("",
                             ok_status=None,
                             data=data,
                             headers=headers)

    def delete(self):
        """Delete template config for specified template name.

        .. __: https://api.go.cd/current/#delete-a-template

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        headers = self._default_headers()

        return self._request(self.name,
                             ok_status=None,
                             data=None,
                             headers=headers,
                             method="DELETE")

    def _default_headers(self):
        return {"Accept": self._accept_header_value,
                "Content-Type": "application/json"}

    @property
    def _accept_header_value(self):
        return "application/vnd.go.cd.v{0}+json".format(self.api_version)

    @staticmethod
    def _json_encode(config):
        return json.dumps(config)



import time
from gocd.api.response import Response
from gocd.api.endpoint import Endpoint
from gocd.api.artifact import Artifact
from gocd.api.stage import Stage

__all__ = ['Pipeline']


class Pipeline(Endpoint):
    base_path = 'go/api/pipelines/{id}'
    id = 'name'
    #: The result of a job/stage has been finalised when these values are set
    final_results = ['Passed', 'Failed']

    def __init__(self, server, name, api_version=1):
        """A wrapper for the `Go pipeline API`__

        .. __: http://api.go.cd/current/#pipelines

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          name (str): The name of the pipeline we're working on
        """
        self.server = server
        self.name = name
        self.api_version = api_version

    def history(self, offset=0):
        """Lists previous instances/runs of the pipeline

        See the `Go pipeline history documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-pipeline-history

        Args:
          offset (int, optional): How many instances to skip for this response.

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get('/history/{offset:d}'.format(offset=offset or 0))

    def release(self):
        """Releases a previously locked pipeline

        See the `Go pipeline release lock documentation`__ for example
        responses.

        .. __: http://api.go.cd/current/#releasing-a-pipeline-lock

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._post('/unlock', headers=self._default_headers())

    #: This is an alias for :meth:`release`
    unlock = release

    def pause(self, reason=''):
        """Pauses the current pipeline

        See the `Go pipeline pause documentation`__ for example responses.

        .. __: http://api.go.cd/current/#pause-a-pipeline

        Args:
          reason (str, optional): The reason the pipeline is being paused.

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._post('/pause', headers=self._default_headers(True), pause_cause=reason)

    def unpause(self):
        """Unpauses the pipeline

        See the `Go pipeline unpause documentation`__ for example responses.

        .. __: http://api.go.cd/current/#unpause-a-pipeline

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._post('/unpause', headers=self._default_headers())

    def status(self):
        """Returns the current status of this pipeline

        See the `Go pipeline status documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-pipeline-status

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get('/status')

    def instance(self, counter=None):
        """Returns all the information regarding a specific pipeline run

        See the `Go pipeline instance documentation`__ for examples.

        .. __: http://api.go.cd/current/#get-pipeline-instance

        Args:
          counter (int): The pipeline instance to fetch.
            If falsey returns the latest pipeline instance from :meth:`history`.

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        if not counter:
            history = self.history()
            if not history:
                return history
            else:
                return Response._from_json(history['pipelines'][0])

        return self._get('/instance/{counter:d}'.format(counter=counter))

    def schedule(self, variables=None, secure_variables=None, materials=None,
                 return_new_instance=False, backoff_time=1.0):
        """Schedule a pipeline run

        Aliased as :meth:`run`, :meth:`schedule`, and :meth:`trigger`.

        Args:
          variables (dict, optional): Variables to set/override
          secure_variables (dict, optional): Secure variables to set/override
          materials (dict, optional): Material revisions to be used for
            this pipeline run. The exact format for this is a bit iffy,
            have a look at the official
            `Go pipeline scheduling documentation`__ or inspect a call
            from triggering manually in the UI.
          return_new_instance (bool): Returns a :meth:`history` compatible
            response for the newly scheduled instance. This is primarily so
            users easily can get the new instance number. **Note:** This is done
            in a very naive way, it just checks that the instance number is
            higher than before the pipeline was triggered.
          backoff_time (float): How long between each check for
            :arg:`return_new_instance`.

         .. __: http://api.go.cd/current/#scheduling-pipelines

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        scheduling_args = dict(
            variables=variables,
            secure_variables=secure_variables,
            material_fingerprint=materials,
            headers=self._default_headers(),
        )

        scheduling_args = dict((k, v) for k, v in scheduling_args.items() if v is not None)

        # TODO: Replace this with whatever is the official way as soon as gocd#990 is fixed.
        # https://github.com/gocd/gocd/issues/990
        if return_new_instance:
            pipelines = self.history()['pipelines']
            if len(pipelines) == 0:
                last_run = None
            else:
                last_run = pipelines[0]['counter']
            response = self._post('/schedule', ok_status=202, **scheduling_args)
            if not response:
                return response

            max_tries = 10
            while max_tries > 0:
                current = self.instance()
                if not last_run and current:
                    return current
                elif last_run and current['counter'] > last_run:
                    return current
                else:
                    time.sleep(backoff_time)
                    max_tries -= 1

            # I can't come up with a scenario in testing where this would happen, but it seems
            # better than returning None.
            return response
        else:
            return self._post('/schedule', ok_status=202, **scheduling_args)

    #: This is an alias for :meth:`schedule`
    run = schedule
    #: This is an alias for :meth:`schedule`
    trigger = schedule

    def artifact(self, counter, stage, job, stage_counter=1):
        """Helper to instantiate an :class:`gocd.api.artifact.Artifact` object

        Args:
          counter (int): The pipeline counter to get the artifact for
          stage: Stage name
          job: Job name
          stage_counter: Defaults to 1

        Returns:
          Artifact: :class:`gocd.api.artifact.Artifact` object
        """
        return Artifact(self.server, self.name, counter, stage, job, stage_counter)

    # TODO: It would be nice if this could stream the output as it happens.
    # Currently it's built with the assumption that this is done after all output has finished.
    def console_output(self, instance=None):
        """Yields the output and metadata from all jobs in the pipeline

        Args:
          instance: The result of a :meth:`instance` call, if not supplied
            the latest of the pipeline will be used.

        Yields:
          tuple: (metadata (dict), output (str)).

          metadata contains:
            - pipeline
            - pipeline_counter
            - stage
            - stage_counter
            - job
            - job_result
        """
        if instance is None:
            instance = self.instance()

        for stage in instance['stages']:
            for job in stage['jobs']:
                if job['result'] not in self.final_results:
                    continue

                artifact = self.artifact(
                    instance['counter'],
                    stage['name'],
                    job['name'],
                    stage['counter']
                )
                output = artifact.get('cruise-output/console.log')

                yield (
                    {
                        'pipeline': self.name,
                        'pipeline_counter': instance['counter'],
                        'stage': stage['name'],
                        'stage_counter': stage['counter'],
                        'job': job['name'],
                        'job_result': job['result'],
                    },
                    output.body
                )

    def stage(self, name, pipeline_counter=None):
        """Helper to instantiate a :class:`gocd.api.stage.Stage` object

        Args:
            name: The name of the stage
            pipeline_counter:

        Returns:

        """
        return Stage(
            self.server,
            pipeline_name=self.name,
            stage_name=name,
            pipeline_counter=pipeline_counter,
        )

    def _default_headers(self, has_data=False):
        if has_data:
            return {
                "Accept": self._accept_header_value,
                "Content-Type": "application/json"
            }
        return {
            "Accept": self._accept_header_value,
            "X-GoCD-Confirm": True
        }

    @property
    def _accept_header_value(self):
        return "application/vnd.go.cd.v{0}+json".format(self.api_version)



import json
from gocd.api.endpoint import Endpoint

__all__ = ['PipelineConfig']


class PipelineConfig(Endpoint):
    base_path = 'go/api/admin/pipelines'
    id = 'name'
    #: The result of a job/stage has been finalised when these values are set
    final_results = ['Passed', 'Failed']

    def __init__(self, server, name, api_version=6):
        """A wrapper for the `Go pipeline config API`__

        .. __: https://api.go.cd/current/#pipeline-config

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          name (str): The name of the pipeline we're working on
        """
        self.server = server
        self.name = name
        self.api_version = api_version

    def get(self):
        """Gets pipeline config for specified pipeline name.

        See `The pipeline config object`__ for example responses.

        .. __: https://api.go.cd/current/#the-pipeline-config-object

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get(self.name, headers={"Accept": self._accept_header_value})

    def edit(self, config, etag):
        """Update pipeline config for specified pipeline name.

        .. __: https://api.go.cd/current/#edit-pipeline-config

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        data = self._json_encode(config)
        headers = self._default_headers()

        if etag is not None:
            headers["If-Match"] = etag

        return self._request(self.name,
                             ok_status=None,
                             data=data,
                             headers=headers,
                             method="PUT")

    def create(self, config):
        """Update pipeline config for specified pipeline name.

        .. __: https://api.go.cd/current/#edit-pipeline-config

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        assert config["name"] == self.name, "Given config is not for this pipeline"
        assert "group" in config, "Given config has no group"

        data = self._json_encode(config)
        headers = self._default_headers()

        return self._request("",
                             ok_status=None,
                             data=data,
                             headers=headers)

    def _default_headers(self):
        return {"Accept": self._accept_header_value,
                "Content-Type": "application/json"}

    @property
    def _accept_header_value(self):
        return "application/vnd.go.cd.v{0}+json".format(self.api_version)

    @staticmethod
    def _json_encode(config):
        return json.dumps(config)



import json
from gocd.api.endpoint import Endpoint

__all__ = ['PluggableSCM']


class PluggableSCM(Endpoint):
    base_path = 'go/api/admin/scms'
    id = 'name'

    def __init__(self, server, name=""):
        """A wrapper for the `Go pluggable SCM API`__

        .. __: https://api.go.cd/current/#scms

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          name (str): The name of the SCM material
        """
        self.server = server
        self.name = name

    def list(self):
        """Lists all available pluggable scm materials,
           these are materials that are present in the in cruise-config.xml.

        See the `Go pluggable SCM documentation`__ for example responses.

        .. __: https://api.go.cd/current/#get-all-pluggable-scm-materials

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get("", headers={"Accept": self._accept_header_value})

    def get(self):
        """Gets SCM material for specified material name

        See `The global scm config object`__ for example responses.

        .. __: https://api.go.cd/current/#the-global-scm-config-object

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get(self.name, headers={"Accept": self._accept_header_value})

    def edit(self, config, etag):
        """Update SCM material for specified material  name.

        .. __: https://api.go.cd/current/#update-pluggable-scm-object

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        data = self._json_encode(config)
        headers = self._default_headers()

        if etag is not None:
            headers["If-Match"] = etag

        return self._request(self.name,
                             ok_status=None,
                             data=data,
                             headers=headers,
                             method="PUT")

    def create(self, config):
        """Create a global SCM object

        .. __: https://api.go.cd/current/#create-a-scm-object

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """

        assert config["name"] == self.name, "SCM configuration name doesn't match"
        assert "configuration" in config, "SCM has no configuration"

        data = self._json_encode(config)
        headers = self._default_headers()

        return self._request("",
                             ok_status=None,
                             data=data,
                             headers=headers,
                             method="POST")

    def _default_headers(self):
        return {"Accept": self._accept_header_value,
                "Content-Type": "application/json"}

    @property
    def _accept_header_value(self):
        return "application/vnd.go.cd.v1+json"

    @staticmethod
    def _json_encode(config):
        return json.dumps(config)



from gocd.api.endpoint import Endpoint
import time


class Artifact(Endpoint):
    base_path = 'go/files/{pipeline}/{counter}/{stage}/{stage_counter}/{job}'

    def __init__(self, server, pipeline, counter, stage, job, stage_counter=1):
        """A wrapper for the `Go artifact API`__

        .. __: http://api.go.cd/current/#artifacts

        Args:
          server (Server): A configured instance of
            :class:gocd.server.Server
          pipeline (str): The name of the pipeline to work with
          counter (int): The counter of the pipeline to work with
          stage (str): The name of the stage to work with
          job (str): The name of the job to work with
          stage_counter (int): The counter of the stage to work with, defaults to 1
        """
        self.server = server
        self.pipeline = pipeline
        self.counter = counter
        self.stage = stage
        self.job = job
        self.stage_counter = stage_counter

        self._base_path = self.base_path.format(
            pipeline=self.pipeline,
            counter=self.counter,
            stage=self.stage,
            stage_counter=self.stage_counter,
            job=self.job
        )

    def list(self):
        """Lists all available artifacts in this job.

        See the `Go artifact list documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-all-artifacts

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get('.json')

    def get(self, path_to_file):
        """Gets an artifact directory by its path.

        See the `Go artifact file documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-artifact-file

        Args:
          path_to_file (str): The path to file to get. It can be nested eg
            ``dist/foobar-widgets-1.2.0.jar``

        Returns:
          Response: :class:`gocd.api.response.Response` object
        """
        return self._get(path_to_file)

    def get_directory(self, path_to_directory, timeout=30, backoff=0.4, max_wait=4):
        """Gets an artifact directory by its path.

        See the `Go artifact directory documentation`__ for example responses.

        .. __: http://api.go.cd/current/#get-artifact-directory

        .. note::
          Getting a directory relies on Go creating a zip file of the
          directory in question. Because of this Go will zip the file in
          the background and return a 202 Accepted response. It's then up
          to the client to check again later and get the final file.

          To work with normal assumptions this :meth:`get_directory` will
          retry itself up to ``timeout`` seconds to get a 200 response to
          return. At that point it will then return the response as is, no
          matter whether it's still 202 or 200. The retry is done with an
          exponential backoff with a max value between retries. See the
          ``backoff`` and ``max_wait`` variables.

          If you want to handle the retry logic yourself then use :meth:`get`
          and add '.zip' as a suffix on the directory.

        Args:
          path_to_directory (str): The path to the directory to get.
            It can be nested eg ``target/dist.zip``
          timeout (int): How many seconds we will wait in total for a
            successful response from Go when we're receiving 202
          backoff (float): The initial value used for backoff, raises
            exponentially until it reaches ``max_wait``
          max_wait (int): The max time between retries

        Returns:
          Response: :class:`gocd.api.response.Response` object
            A successful response is a zip-file.
        """
        response = None
        started_at = None
        time_elapsed = 0

        i = 0
        while time_elapsed < timeout:
            response = self._get('{0}.zip'.format(path_to_directory))

            if response:
                break
            else:
                if started_at is None:
                    started_at = time.time()

                time.sleep(min(backoff * (2 ** i), max_wait))
                i += 1
                time_elapsed = time.time() - started_at

        return response



import urllib

import pytest

from gocd.vendor.multidimensional_urlencode import urlencode


def test_basic():
    """Verify that urlencode works with four levels."""
    d = {"a": {"b": {"c": "d"}}}
    expected = urllib.quote("a[b][c]=d", safe="=/&")
    assert urlencode(d) == expected


def test_key_types():
    """Verify that urlencode works with key type 'int'."""
    d = {1: {2: {3: 4}}}
    expected = urllib.quote("1[2][3]=4", safe="=/&")
    assert urlencode(d) == expected


def test_two():
    """Verify that urlencode works with two params."""
    d = {'a': 'b', 'c': {'d': 'e'}}
    expected = urllib.quote("a=b&c[d]=e", safe="=/&")
    assert urlencode(d) == expected


def test_with_list_value():
    """Verify that urlencode works with list value."""
    d = {'a': {"b": [1, 2, 3]}}
    expected = "a[b][]=1&a[b][]=2&a[b][]=3"
    assert urllib.unquote(urlencode(d)) == expected


def test_with_non_dict():
    """Verify that we raise an exception when passing a non-dict."""
    with pytest.raises(TypeError):
        urlencode("e")
